@@ -18,6 +18,7 @@ import io
 import math
 import sys
 import time
+import traceback
 import types
 import unittest
 import warnings
@@ -60,6 +61,7 @@ from hypothesis.errors import (
     Flaky,
     Found,
     HypothesisDeprecationWarning,
+    HypothesisException,
     HypothesisWarning,
     InvalidArgument,
     NoSuchExample,
@@ -77,15 +79,18 @@ from hypothesis.internal.compat import (
 )
 from hypothesis.internal.conjecture.data import ConjectureData, Status
 from hypothesis.internal.conjecture.engine import BUFFER_SIZE, ConjectureRunner
-from hypothesis.internal.conjecture.junkdrawer import ensure_free_stackframes
+from hypothesis.internal.conjecture.junkdrawer import (
+    ensure_free_stackframes,
+    gc_cumulative_time,
+)
 from hypothesis.internal.conjecture.shrinker import sort_key
 from hypothesis.internal.entropy import deterministic_PRNG
 from hypothesis.internal.escalation import (
     InterestingOrigin,
     current_pytest_item,
-    escalate_hypothesis_internal_error,
     format_exception,
     get_trimmed_traceback,
+    is_hypothesis_file,
 )
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.internal.observability import (
@@ -820,21 +825,34 @@ class StateForActualGivenExecution:
         self._string_repr = ""
         text_repr = None
         if self.settings.deadline is None and not TESTCASE_CALLBACKS:
-            test = self.test
+
+            @proxies(self.test)
+            def test(*args, **kwargs):
+                with ensure_free_stackframes():
+                    return self.test(*args, **kwargs)
+
         else:
 
             @proxies(self.test)
             def test(*args, **kwargs):
                 arg_drawtime = math.fsum(data.draw_times.values())
+                arg_stateful = math.fsum(data._stateful_run_times.values())
+                arg_gctime = gc_cumulative_time()
                 start = time.perf_counter()
                 try:
-                    result = self.test(*args, **kwargs)
+                    with ensure_free_stackframes():
+                        result = self.test(*args, **kwargs)
                 finally:
                     finish = time.perf_counter()
                     in_drawtime = math.fsum(data.draw_times.values()) - arg_drawtime
-                    runtime = datetime.timedelta(seconds=finish - start - in_drawtime)
+                    in_stateful = (
+                        math.fsum(data._stateful_run_times.values()) - arg_stateful
+                    )
+                    in_gctime = gc_cumulative_time() - arg_gctime
+                    runtime = finish - start - in_drawtime - in_stateful - in_gctime
                     self._timing_features = {
-                        "execute:test": finish - start - in_drawtime,
+                        "execute:test": runtime,
+                        "overall:gc": in_gctime,
                         **data.draw_times,
                         **data._stateful_run_times,
                     }
@@ -842,8 +860,10 @@ class StateForActualGivenExecution:
                 if (current_deadline := self.settings.deadline) is not None:
                     if not is_final:
                         current_deadline = (current_deadline // 4) * 5
-                    if runtime >= current_deadline:
-                        raise DeadlineExceeded(runtime, self.settings.deadline)
+                    if runtime >= current_deadline.total_seconds():
+                        raise DeadlineExceeded(
+                            datetime.timedelta(seconds=runtime), self.settings.deadline
+                        )
                 return result
 
         def run(data):
@@ -1053,7 +1073,11 @@ class StateForActualGivenExecution:
         except failure_exceptions_to_catch() as e:
             # If the error was raised by Hypothesis-internal code, re-raise it
             # as a fatal error instead of treating it as a test failure.
-            escalate_hypothesis_internal_error()
+            filepath = traceback.extract_tb(e.__traceback__)[-1][0]
+            if is_hypothesis_file(filepath) and not isinstance(
+                e, (HypothesisException, StopTest, UnsatisfiedAssumption)
+            ):
+                raise
 
             if data.frozen:
                 # This can happen if an error occurred in a finally
@@ -1086,10 +1110,11 @@ class StateForActualGivenExecution:
             if TESTCASE_CALLBACKS:
                 if runner := getattr(self, "_runner", None):
                     phase = runner._current_phase
-                elif self.failed_normally or self.failed_due_to_deadline:
-                    phase = "shrink"
                 else:  # pragma: no cover  # in case of messing with internals
-                    phase = "unknown"
+                    if self.failed_normally or self.failed_due_to_deadline:
+                        phase = "shrink"
+                    else:
+                        phase = "unknown"
                 backend_desc = f", using backend={self.settings.backend!r}" * (
                     self.settings.backend != "hypothesis"
                     and not getattr(runner, "_switch_to_hypothesis_provider", False)
